from typing import TypeVar

from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from .compat import CachingProtocolMeta

T_contra = TypeVar("T_contra", contravariant=True)
U_co = TypeVar("U_co", covariant=True)


@runtime_checkable
class _SupportsLt(Protocol[T_contra]):
    def __lt__(self, other: T_contra) -> bool:
        ...


class SupportsLt(
    _SupportsLt[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsLe(Protocol[T_contra]):
    def __le__(self, other: T_contra) -> bool:
        ...


class SupportsLe(
    _SupportsLe[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsGt(Protocol[T_contra]):
    def __gt__(self, other: T_contra) -> bool:
        ...


class SupportsGt(
    _SupportsGt[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsGe(Protocol[T_contra]):
    def __ge__(self, other: T_contra) -> bool:
        ...


class SupportsGe(
    _SupportsGe[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsEq(Protocol):
    def __eq__(self, other: object) -> bool:
        ...


class SupportsEq(Protocol, metaclass=CachingProtocolMeta):
    ...


@runtime_checkable
class _Comparable(
    SupportsLt[T_contra],
    SupportsLe[T_contra],
    SupportsGt[T_contra],
    SupportsGe[T_contra],
    SupportsEq,
    Protocol[T_contra],
):
    ...


class Comparable(
    _Comparable[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsLeGe(SupportsLe[T_contra], SupportsGe[T_contra], Protocol[T_contra]):
    ...


class SupportsLeGe(
    _SupportsLeGe[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsLeGt(SupportsLe[T_contra], SupportsGt[T_contra], Protocol[T_contra]):
    ...


class SupportsLeGt(
    _SupportsLeGt[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsLtGe(SupportsLt[T_contra], SupportsGe[T_contra], Protocol[T_contra]):
    ...


class SupportsLtGe(
    _SupportsLtGe[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


class _SupportsLtGt(SupportsLt[T_contra], SupportsGt[T_contra], Protocol[T_contra]):
    ...


class SupportsLtGt(
    _SupportsLtGt[T_contra],
    Protocol[T_contra],
    metaclass=CachingProtocolMeta,
):
    ...


@runtime_checkable
class _SupportsMod(Protocol[T_contra, U_co]):
    def __mod__(self, other: T_contra) -> U_co:
        ...


class SupportsMod(
    _SupportsMod[T_contra, U_co],
    Protocol[T_contra, U_co],
    metaclass=CachingProtocolMeta,
):
    ...



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Iterable
from typing import Sequence
from typing import TypeVar
from typing import cast

from typing_extensions import Final
from typing_extensions import TypeAlias
from typing_extensions import get_args

from ._utils.misc import is_union
from .errors import BoundError
from .predicates.boolean import all_of
from .predicates.generic import of_complex_type
from .predicates.generic import of_type

__all__ = ("get_bound_parser", "parse_str", "Parser")

T = TypeVar("T", covariant=True)
Parser: TypeAlias = Callable[[object], T]


def display_bound(bound: Any) -> str:
    if isinstance(bound, Iterable):
        return f"Intersection[{', '.join(display_bound(part) for part in bound)}]"
    if is_union(bound):
        return (
            f"typing.Union["
            f"{', '.join(display_bound(part) for part in get_args(bound))}"
            f"]"
        )
    return str(getattr(bound, "__name__", bound))


def get_bound_parser(bound: type[T] | Any) -> Parser[T]:
    within_bound = (
        # Interpret sequence as intersection
        all_of(of_type(t) for t in bound)
        if isinstance(bound, Sequence)
        else of_complex_type(bound)
    )

    def parser(instance: object) -> T:
        if not within_bound(instance):
            raise BoundError(
                f"Value is not within bound of {display_bound(bound)!r}: {instance!r}"
            )
        return cast(T, instance)

    return parser


parse_str: Final[Callable[[object], str]] = get_bound_parser(str)



from typing import TYPE_CHECKING

# Workaround for mypy crash, see:
#   https://github.com/antonagestam/phantom-types/issues/243
if TYPE_CHECKING:

    class CachingProtocolMeta(type):
        ...

else:
    try:
        from numerary.protocol import CachingProtocolMeta
    except ImportError:
        from numerary.types import CachingProtocolMeta


__all__ = ("CachingProtocolMeta",)



"""
Types for narrowing on the builtin datetime types.

These types can be used without installing any extra dependencies, however, to parse
strings, python-dateutil must be installed or a
:py:class:`phantom.errors.MissingDependency` error will be raised when calling parse.

You can install python-dateutil by using the ``[dateutil]`` or ``[all]`` extras.
"""
from __future__ import annotations

import datetime

from . import Phantom
from .bounds import parse_str
from .errors import MissingDependency
from .predicates.datetime import is_tz_aware
from .predicates.datetime import is_tz_naive
from .schema import Schema

try:
    import dateutil.parser

    parse_datetime_str = dateutil.parser.parse
    DateutilParseError = dateutil.parser.ParserError
except ImportError as e:
    exception = e

    def parse_datetime_str(
        *_: object,
        **__: object,
    ) -> datetime.datetime:
        raise MissingDependency(
            "python-dateutil needs to be installed to use this type for parsing. It "
            "can be installed with the phantom-types[dateutil] extra."
        ) from exception

    class DateutilParseError(Exception):  # type: ignore[no-redef]
        ...


__all__ = ("TZAware", "TZNaive")


def parse_datetime(value: object) -> datetime.datetime:
    if isinstance(value, datetime.datetime):
        return value
    str_value = parse_str(value)
    try:
        return parse_datetime_str(str_value)
    except DateutilParseError as exc:
        raise TypeError("Could not parse datetime from given string") from exc


class TZAware(datetime.datetime, Phantom, predicate=is_tz_aware):
    """
    A type for helping ensure that ``datetime`` objects are always timezone aware.

    >>> isinstance(datetime.datetime.now(), TZAware)
    False
    >>> isinstance(datetime.datetime.now(tz=datetime.timezone.utc), TZAware)
    True
    """

    # A property of being aware is (dt.tzinfo != None), so we can safely narrow this
    # attribute to not include None.
    tzinfo: datetime.tzinfo

    @classmethod
    def parse(cls, instance: object) -> TZAware:
        return super().parse(parse_datetime(instance))

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A date-time with timezone data.",
        }


class TZNaive(datetime.datetime, Phantom, predicate=is_tz_naive):
    """
    >>> isinstance(datetime.datetime.now(), TZNaive)
    True
    >>> isinstance(datetime.datetime.now(tz=datetime.timezone.utc), TZNaive)
    False
    """

    @classmethod
    def parse(cls, instance: object) -> TZNaive:
        return super().parse(parse_datetime(instance))

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A date-time without timezone data.",
            "format": "date-time-naive",
        }



"""
Use ``Phantom`` to create arbitrary phantom types using boolean predicates.

.. code-block:: python

    from phantom import Phantom


    def is_big(value: int) -> bool:
        return value > 5


    class Big(int, Phantom, predicate=is_big):
        ...


    assert isinstance(10, Big)  # this passes
"""
from ._base import Phantom
from ._base import PhantomBase
from ._base import PhantomMeta
from .bounds import get_bound_parser
from .errors import BoundError
from .predicates import Predicate

__version__ = "1.1.0"
__all__ = (
    "BoundError",
    "Phantom",
    "PhantomBase",
    "PhantomMeta",
    "get_bound_parser",
    "Predicate",
)



"""
This module provides a single type: :py:class:`SequenceNotStr`. This type is equivalent
to :py:class:`typing.Sequence` except it excludes values of type :py:class:`str` and
:py:class:`bytes` from the set of valid instances. This can be useful when you want to
eliminate the easy mistake of forgetting to wrap a string value in a containing
sequence.
"""

from typing import Generic
from typing import Sequence
from typing import TypeVar

from . import Phantom
from .predicates import boolean
from .predicates.generic import of_type

__all__ = ("SequenceNotStr",)

T = TypeVar("T")


class SequenceNotStr(
    Sequence[T],
    Phantom,
    Generic[T],
    # Note: We don't eliminate mutable types here like in PhantomSized. This is because
    # the property of not being a str cannot change by mutation, so this specific
    # phantom type is safe to use with mutable types.
    predicate=boolean.negate(of_type((str, bytes))),
):
    ...



"""
Functions that create new predicates that succeed when their argument is strictly or non
strictly between the upper and lower bounds. There are corresponding phantom types that
use these predicates in :py:mod:`phantom.interval`.
"""

from typing import TypeVar

from phantom._utils.types import SupportsLeGe
from phantom._utils.types import SupportsLeGt
from phantom._utils.types import SupportsLtGe
from phantom._utils.types import SupportsLtGt

from ._base import Predicate
from ._utils import bind_name

T = TypeVar("T")


def exclusive(low: T, high: T) -> Predicate[SupportsLtGt[T]]:
    """
    Create a predicate that succeeds when its argument is in the range ``(low, high)``.
    """

    @bind_name(exclusive, low, high)
    def check(value: SupportsLtGt[T]) -> bool:
        return low < value < high

    return check


def exclusive_inclusive(low: T, high: T) -> Predicate[SupportsLeGt[T]]:
    """
    Create a predicate that succeeds when its argument is in the range ``(low, high]``.
    """

    @bind_name(exclusive_inclusive, low, high)
    def check(value: SupportsLeGt[T]) -> bool:
        return low < value <= high

    return check


def inclusive_exclusive(low: T, high: T) -> Predicate[SupportsLtGe[T]]:
    """
    Create a predicate that succeeds when its argument is in the range ``[low, high)``.
    """

    @bind_name(inclusive_exclusive, low, high)
    def check(value: SupportsLtGe[T]) -> bool:
        return low <= value < high

    return check


def inclusive(low: T, high: T) -> Predicate[SupportsLeGe[T]]:
    """
    Create a predicate that succeeds when its argument is in the range ``[low, high]``.
    """

    @bind_name(inclusive, low, high)
    def check(value: SupportsLeGe[T]) -> bool:
        return low <= value <= high

    return check



"""
Types describing collections with size boundaries. These types should only be used with
immutable collections. There is a naive check that eliminates some of the most common
mutable collections in the instance check. However, a guaranteed check is probably
impossible to implement, so some amount of developer discipline is required.

Sized types are created by subclassing :py:class:`PhantomBound` and providing a minimum,
maximum, or both as the ``min`` and ``max`` class arguments. For instance,
:py:class:`NonEmpty` is implemented using ``min=1``.

This made-up type would describe sized collections with between 5 and 10 ints:

.. code-block:: python

    class SpecificSize(PhantomBound[int], min=5, max=10):
        ...


This example creates a type that accepts strings with 255 or less characters:

.. code-block:: python

    class SizedStr(str, PhantomBound[str], max=255):
        ...

"""
from __future__ import annotations

from typing import Any
from typing import Generic
from typing import Iterable
from typing import Sized
from typing import TypeVar

from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from . import Phantom
from . import PhantomMeta
from . import Predicate
from ._utils.misc import is_not_known_mutable_instance
from .predicates import boolean
from .predicates import collection
from .predicates import interval
from .predicates import numeric
from .schema import Schema

# We attempt to import _ProtocolMeta from typing_extensions to support Python 3.7 but
# fall back the typing module to support Python 3.8+. This is the closest I could find
# to documentation of _ProtocolMeta.
# https://github.com/python/cpython/commit/74d7f76e2c953fbfdb7ce01b7319d91d471cc5ef
try:
    from typing_extensions import _ProtocolMeta  # type: ignore[attr-defined]
except ImportError:
    from typing import _ProtocolMeta

__all__ = (
    "SizedIterable",
    "PhantomSized",
    "PhantomBound",
    "NonEmpty",
    "NonEmptyStr",
    "Empty",
)


T = TypeVar("T", bound=object, covariant=True)


@runtime_checkable
class SizedIterable(Sized, Iterable[T], Protocol[T]):
    """Intersection of :py:class:`typing.Sized` and :py:class:`typing.Iterable`."""


class SizedIterablePhantomMeta(PhantomMeta, _ProtocolMeta):  # type: ignore[misc]
    ...


class PhantomSized(
    Phantom[Sized],
    SizedIterable[T],
    Generic[T],
    metaclass=SizedIterablePhantomMeta,
    bound=SizedIterable,
    abstract=True,
):
    """
    Takes class argument ``len: Predicate[int]``.

    Discouraged in favor of :py:class:`PhantomBound`, which better supports automatic
    schema generation.
    """

    def __init_subclass__(cls, len: Predicate[int], **kwargs: Any) -> None:
        super().__init_subclass__(
            predicate=boolean.both(
                is_not_known_mutable_instance,
                collection.count(len),
            ),
            **kwargs,
        )

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "type": "array",
        }


class UnresolvedBounds(Exception):
    ...


class LSPViolation(Exception):
    ...


class PhantomBound(
    Phantom[Sized],
    SizedIterable[T],
    Generic[T],
    metaclass=SizedIterablePhantomMeta,
    bound=SizedIterable,
    abstract=True,
):
    """Takes class arguments ``min: int``, ``max: int``."""

    __min__: int | None
    __max__: int | None

    def __init_subclass__(  # noqa
        cls,
        min: int | None = None,
        max: int | None = None,
        abstract: bool = False,
        **kwargs: Any,
    ) -> None:
        inherited_min = getattr(cls, "__min__", None)
        inherited_max = getattr(cls, "__max__", None)
        cls.__min__ = inherited_min if min is None else min
        cls.__max__ = inherited_max if max is None else max

        # Note: There's possibly value in generalizing this, to be able to declaratively
        #       describe the relationship between an attribute and its inherited value.
        if (
            cls.__min__ is not None
            and inherited_min is not None
            and cls.__min__ < inherited_min
        ):
            raise LSPViolation(
                f"Cannot set a smaller min than inherited ({cls.__min__} < "
                f"{inherited_min})."
            )

        if (
            cls.__max__ is not None
            and inherited_max is not None
            and cls.__max__ > inherited_max
        ):
            raise LSPViolation(
                f"Cannot set a larger max than inherited ({cls.__max__} > "
                f"{inherited_max})."
            )

        if cls.__min__ is not None and cls.__max__ is not None:
            size = interval.inclusive(cls.__min__, cls.__max__)
        elif cls.__min__ is not None:
            size = numeric.ge(cls.__min__)
        elif cls.__max__ is not None:
            size = numeric.le(cls.__max__)
        elif abstract:
            super().__init_subclass__(abstract=abstract, **kwargs)
            return
        else:
            raise UnresolvedBounds(
                f"Concrete type {cls.__qualname__} must provide either min or max, or "
                f"both."
            )

        super().__init_subclass__(
            predicate=boolean.both(
                is_not_known_mutable_instance,
                collection.count(size),
            ),
            abstract=abstract,
            **kwargs,
        )

    @classmethod
    def __schema__(cls) -> Schema:
        return (
            {
                **super().__schema__(),  # type: ignore[misc]
                "type": "string",
                "minLength": cls.__min__,
                "maxLength": cls.__max__,
            }
            if str in cls.__mro__
            else {
                **super().__schema__(),  # type: ignore[misc]
                "type": "array",
                "minItems": cls.__min__,
                "maxItems": cls.__max__,
            }
        )


class NonEmpty(PhantomBound[T], Generic[T], min=1):
    """A sized collection with at least one item."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A non-empty array.",
        }


class NonEmptyStr(str, NonEmpty[str]):
    """A sized str with at least one character."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A non-empty string.",
        }


class Empty(PhantomBound[T], Generic[T], max=0):
    """A sized collection with exactly zero items."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "An empty array.",
        }
